module heatbath_mod
  implicit none
  private
  public :: DP
  public :: PI
  public ::     new_rand
  public ::  delete_rand
  public :: uniform_rand
  public ::   gauss_rand
  public ::     exp_rand

  integer, parameter :: DP = KIND(1.0d0)
  integer,  save, allocatable :: seed(:)
  integer,  save :: seed_size
  real(DP), save :: PI
  logical,  save :: is_initialized = .false.

contains

subroutine new_rand(iseed)
  implicit none
  integer, intent(in) :: iseed
  call RANDOM_SEED(size=seed_size)
  allocate(seed(seed_size))
  seed(:) = iseed
  call RANDOM_SEED(put=seed)

  PI = 4*atan(1.0_DP)
  is_initialized = .true.

  return
end subroutine

subroutine delete_rand()
  implicit none
  if (allocated(seed)) deallocate(seed)
  is_initialized = .false.
  return
end subroutine

subroutine error_rand(sub_name)
  implicit none
  character(*), intent(in) :: sub_name
  write(*,'(" Error at ",A,". Stop!")')TRIM(ADJUSTL(sub_name))
  stop
end subroutine

function uniform_rand(a,b) result(rand_num)
!
! return a uniform random number in [a,b)
!
  implicit none
  real(DP), intent(in) :: a,b
  real(DP) :: rand_num
  real(DP) :: rr
  if (.not.is_initialized) call error_rand("uniform_rand")
  if ( b < a ) call error_rand("uniform_rand(a,b): b < a")
  call RANDOM_NUMBER(rr)   ! rr in [0,1)
  rand_num = (b-a)*rr+a
  return
end function

function gauss_rand() result(rand_num)
!
! Generate a Gaussina random number (mean=0,variance=1) via Box-Muller method.
!
! P(x) \propt exp(-x**2/2)
!
  implicit none
  real(DP) :: rand_num
  real(DP) :: rr,tt
  if (.not.is_initialized) call error_rand("gauss_rand")
  call RANDOM_NUMBER(rr)   ! rr in [0,1)
  call RANDOM_NUMBER(tt)   ! tt in [0,1)
  rr = 1.0_DP - rr     ! convert to rr in (0,1]

  rand_num = SQRT(-2*LOG(rr))*COS(2*PI*tt)

  return
end function

function exp_rand() result(rand_num)
!
! Generate a random number with exponential distribution [0,infinity)
!
!
  implicit none
  real(DP) :: rand_num
  real(DP) :: rr,tt
  if (.not.is_initialized) call error_rand("exp_rand")
  call RANDOM_NUMBER(rr)   ! rr in [0,1)
  rr = 1.0_DP - rr     ! convert to rr in (0,1]
  rand_num = -LOG(rr)
  return
end function

end module

module two_site_scalar_module
!
! $Id: two_site_scalar_metropolis.f90,v 1.3 2011/03/02 06:29:57 ishikawa Exp $
!
! 2-site scalar
! for ASLFT2011@TATA (2011/03/12-)
!
! Ken-Ichi Ishikawa
!
! NO WARRANTY
!
  use heatbath_mod
  implicit none

contains

subroutine generate_state_with_gaussian(s,s0,var)
!
! generate s(1),s(2) with gaussian distribution
!
! P(s,s0) \propt exp(-(s-s0)**2/2/var)
!
  implicit none
  real(DP), intent(inout) :: s(2)
  real(DP), intent(in) :: s0(2)
  real(DP), intent(in) :: var
  s(1) = gauss_rand()*SQRT(var) + s0(1)
  s(2) = gauss_rand()*SQRT(var) + s0(2)
  return
end subroutine

function hamil(beta,s) result(h)
!
! return action/Hamiltonian value for a state
!
!   beta : Hamiltonian parameter
!      s : 2site scalars
!      h : Hamiltonian value
!
!  h = beta*s(1)*(2) + (s(1)**2 + s(2)**2)/2
!
  implicit none
  real(DP), intent(in) :: beta
  real(DP), intent(in) :: s(2)
  real(DP) :: h
  h = beta*s(1)*s(2) + (s(1)**2 + s(2)**2)*0.5_DP
  return
end function

end module

program two_site_scalar_metropolis
!
! $Id: two_site_scalar_metropolis.f90,v 1.3 2011/03/02 06:29:57 ishikawa Exp $
!
! 2-site scalar model metropolis
! for ASLFT2011@TATA (2011/03/12-)
!
! Ken-Ichi Ishikawa
!
! NO WARRANTY
!
  use two_site_scalar_module
  implicit none
  integer :: NTHERM     ! first NTHERM samples are dropped
  integer :: NSAMPLE    ! measured sample number
  integer :: NSKIP      ! skip measured sample
  integer, allocatable :: seed(:)
  integer :: iseed

  real(DP) :: pacc
  integer :: iacc,itry

  real(DP) :: s0(2)  ! previous state (2-spins)
  real(DP) :: s1(2)  !  current state (2-spins)
  integer :: istep
  real(DP) :: rand_num
  real(DP) :: beta  
  real(DP) :: rho,w0,w1,h0,h1
  real(DP) :: spinave,spincorr
  real(DP) :: var = 1.0_DP
  integer :: iout

  iacc = 0
  itry = 0

  iout=99
  open(iout,file="ISING_PARAM",status='old',form='formatted')
  read(iout,*)beta
  read(iout,*)iseed
  read(iout,*)NTHERM,NSAMPLE,NSKIP
  close(iout)


  !
  ! Set up pseudo-random number generator
  !
  call new_rand(iseed)

  write(*,'("# BETA=",ES24.15)') beta

  write(*,'("# ISEED=",I10," NTHERM=",I10," NSAMPLE=",I10," NSKIP=",I10)') &
 &           iseed,NTHERM,NSAMPLE,NSKIP

  write(*,'("# sample#",12X,"s(1)",20X,"s(2)", &
 &            18X,"spin ave",16X,"spin corr")')

  !
  ! Generate initial state with Gaussian distribution with (mean = 0,variance = var)
  ! 
  s0(:) = 0.0_DP
  call generate_state_with_gaussian(s0,s0,var)

  do istep=1,NTHERM+NSAMPLE*NSKIP

    !
    ! Generate a candidate state with Gaussian distribution with (mean = s0,variance = var)
    !
    call generate_state_with_gaussian(s1,s0,var)

    !
    ! Compute Metropolis test weight
    !
    h0 = hamil(beta,s0)
    h1 = hamil(beta,s1)
    w0 = exp(-h0)
    w1 = exp(-h1)
    rho = MIN(1.0_DP,w1/w0)

    !
    ! Metropolis test Accept Reject step
    !
    call RANDOM_NUMBER(rand_num)
    itry = itry + 1
    if (rand_num <= rho) then

      ! accept s1 as the new state

      iacc = iacc + 1
      continue

    else

      ! reject s1. s0 is the new state

      s1(:) = s0(:)

    endif

    if (istep > NTHERM .and. mod(istep,NSKIP) == 0) then
      !
      ! store current state in the ensemble
      ! and measure observables
      !

      spinave = (s1(1) + s1(2))*0.5_DP
      spincorr = s1(1)*s1(2)

      write(*,'(I10,SP,2ES24.15,2ES24.15)')&
  &             istep,s1(1),s1(2), spinave, spincorr

    endif

    !
    ! shift history
    !
    s0(:) = s1(:)

  enddo

  pacc = REAL(iacc,kind=DP)/itry
  write(*,'("# Metropolis test statistics.")')
  write(*,'("# var=",ES14.6," itry=",I10," iacc=",I10," Pacc=",F10.6)') &
 &   var,itry,iacc,pacc*100

  call delete_rand()
  stop
end program
